@@ -4,8 +4,6 @@
 
 #include "system.h"
 
-#include <rpm/rpmcli.h>		/* XXX rpmcliPackagesTotal */
-
 #include <rpm/rpmlib.h>		/* rpmVersionCompare, rpmlib provides */
 #include <rpm/rpmtag.h>
 #include <rpm/rpmlog.h>
@@ -49,6 +47,37 @@ static rpmds rpmlibP = NULL;
 #undef HASHTYPE
 #undef HASHKEYTYPE
 
+/**
+ * Check for supported payload format in header.
+ * @param h		header to check
+ * @return		RPMRC_OK if supported, RPMRC_FAIL otherwise
+ */
+static rpmRC headerCheckPayloadFormat(Header h) {
+    rpmRC rc = RPMRC_OK;
+    const char *payloadfmt = headerGetString(h, RPMTAG_PAYLOADFORMAT);
+    /* 
+     * XXX Ugh, rpm 3.x packages don't have payload format tag. Instead
+     * of blindly allowing, should check somehow (HDRID existence or... ?)
+     */
+    if (!payloadfmt) return rc;
+
+    if (!rstreq(payloadfmt, "cpio")) {
+        char *nevra = headerGetAsString(h, RPMTAG_NEVRA);
+        if (payloadfmt && rstreq(payloadfmt, "drpm")) {
+            rpmlog(RPMLOG_ERR,
+                     _("%s is a Delta RPM and cannot be directly installed\n"),
+                     nevra);
+        } else {
+            rpmlog(RPMLOG_ERR, 
+                     _("Unsupported payload (%s) in package %s\n"),
+                     payloadfmt ? payloadfmt : "none", nevra);
+        } 
+        free(nevra);
+	rc = RPMRC_FAIL;
+    }
+    return rc;
+}
+
 /**
  * Add removed package instance to ordered transaction set.
  * @param ts		transaction set
@@ -70,6 +99,10 @@ static int removePackage(rpmts ts, Header h, rpmte depends)
         return 0;
     }
 
+    p = rpmteNew(ts, h, TR_REMOVED, NULL, NULL);
+    if (p == NULL)
+	return 1;
+
     intHashAddEntry(tsmem->removedPackages, dboffset);
 
     if (tsmem->orderCount >= tsmem->orderAlloced) {
@@ -77,7 +110,6 @@ static int removePackage(rpmts ts, Header h, rpmte depends)
 	tsmem->order = xrealloc(tsmem->order, sizeof(*tsmem->order) * tsmem->orderAlloced);
     }
 
-    p = rpmteNew(ts, h, TR_REMOVED, NULL, NULL);
     rpmteSetDependsOn(p, depends);
 
     tsmem->order[tsmem->orderCount] = p;
@@ -86,24 +118,39 @@ static int removePackage(rpmts ts, Header h, rpmte depends)
     return 0;
 }
 
-/* Return rpmdb iterator with removals pruned out */
-static rpmdbMatchIterator rpmtsPrunedIterator(rpmts ts, rpmDbiTagVal tag, const char * key)
+/* Return rpmdb iterator with removals optionally pruned out */
+static rpmdbMatchIterator rpmtsPrunedIterator(rpmts ts, rpmDbiTagVal tag,
+					      const char * key, int prune)
 {
     rpmdbMatchIterator mi = rpmtsInitIterator(ts, tag, key, 0);
-    tsMembers tsmem = rpmtsMembers(ts);
-    rpmdbPruneIterator(mi, tsmem->removedPackages);
+    if (prune) {
+	tsMembers tsmem = rpmtsMembers(ts);
+	rpmdbPruneIterator(mi, tsmem->removedPackages);
+    }
     return mi;
 }
 
-#define skipColor(_tscolor, _color, _ocolor) \
-	((_tscolor) && (_color) && (_ocolor) && !((_color) & (_ocolor)))
+/**
+ * Decides whether to skip a package upgrade/obsoletion on TE color.
+ *
+ * @param tscolor	color of this transaction
+ * @param color 	color of this transaction element
+ * @param ocolor 	header color of the upgraded/obsoleted package
+ *
+ * @return		non-zero if the package should be skipped
+ */
+static int skipColor(rpm_color_t tscolor, rpm_color_t color, rpm_color_t ocolor)
+{
+    return tscolor && color && ocolor && !(color & ocolor);
+}
 
 /* Add erase elements for older packages of same color (if any). */
-static void addUpgradeErasures(rpmts ts, tsMembers tsmem, rpm_color_t tscolor,
+static int addUpgradeErasures(rpmts ts, rpm_color_t tscolor,
 				rpmte p, rpm_color_t hcolor, Header h)
 {
     Header oh;
     rpmdbMatchIterator mi = rpmtsInitIterator(ts, RPMDBI_NAME, rpmteN(p), 0);
+    int rc = 0;
 
     while((oh = rpmdbNextIterator(mi)) != NULL) {
 	/* Ignore colored packages not in our rainbow. */
@@ -114,53 +161,62 @@ static void addUpgradeErasures(rpmts ts, tsMembers tsmem, rpm_color_t tscolor,
 	if (rpmVersionCompare(h, oh) == 0)
 	    continue;
 
-	removePackage(ts, oh, p);
+	if (removePackage(ts, oh, p)) {
+	    rc = 1;
+	    break;
+	}
     }
-    mi = rpmdbFreeIterator(mi);
+    rpmdbFreeIterator(mi);
+    return rc;
 }
 
 /* Add erase elements for obsoleted packages of same color (if any). */
-static void addObsoleteErasures(rpmts ts, tsMembers tsmem, rpm_color_t tscolor,
-				rpmte p, rpm_color_t hcolor)
+static int addObsoleteErasures(rpmts ts, rpm_color_t tscolor, rpmte p)
 {
     rpmds obsoletes = rpmdsInit(rpmteDS(p, RPMTAG_OBSOLETENAME));
     Header oh;
+    int rc = 0;
 
-    while (rpmdsNext(obsoletes) >= 0) {
+    while (rpmdsNext(obsoletes) >= 0 && rc == 0) {
 	const char * Name;
 	rpmdbMatchIterator mi = NULL;
 
 	if ((Name = rpmdsN(obsoletes)) == NULL)
 	    continue;	/* XXX can't happen */
 
-	/* XXX avoid self-obsoleting packages. */
-	if (rstreq(rpmteN(p), Name))
-	    continue;
-
-	mi = rpmtsPrunedIterator(ts, RPMDBI_NAME, Name);
+	mi = rpmtsPrunedIterator(ts, RPMDBI_NAME, Name, 1);
 
 	while((oh = rpmdbNextIterator(mi)) != NULL) {
-	    /* Ignore colored packages not in our rainbow. */
-	    if (skipColor(tscolor, hcolor, 
-			  headerGetNumber(oh, RPMTAG_HEADERCOLOR)))
+	    const char *oarch = headerGetString(oh, RPMTAG_ARCH);
+
+	    /* avoid self-obsoleting packages */
+	    if (rstreq(rpmteN(p), Name) && rstreq(rpmteA(p), oarch)) {
+		char * ohNEVRA = headerGetAsString(oh, RPMTAG_NEVRA);
+		rpmlog(RPMLOG_DEBUG, "  Not obsoleting: %s\n", ohNEVRA);
+		free(ohNEVRA);
 		continue;
+	    }
 
 	    /*
 	     * Rpm prior to 3.0.3 does not have versioned obsoletes.
 	     * If no obsoletes version info is available, match all names.
 	     */
 	    if (rpmdsEVR(obsoletes) == NULL
-		     || rpmdsAnyMatchesDep(oh, obsoletes, _rpmds_nopromote)) {
+		|| rpmdsNVRMatchesDep(oh, obsoletes, _rpmds_nopromote)) {
 		char * ohNEVRA = headerGetAsString(oh, RPMTAG_NEVRA);
 		rpmlog(RPMLOG_DEBUG, "  Obsoletes: %s\t\terases %s\n",
 			rpmdsDNEVR(obsoletes)+2, ohNEVRA);
-		ohNEVRA = _free(ohNEVRA);
+		free(ohNEVRA);
 
-		removePackage(ts, oh, p);
+		if (removePackage(ts, oh, p)) {
+		    rc = 1;
+		    break;
+		}
 	    }
 	}
-	mi = rpmdbFreeIterator(mi);
+	rpmdbFreeIterator(mi);
     }
+    return rc;
 }
 
 /*
@@ -272,6 +328,22 @@ int rpmtsAddInstallElement(rpmts ts, Header h,
 	goto exit;
     }
 
+    /* Source packages are never "upgraded" */
+    if (isSource)
+	upgrade = 0;
+
+    /* Do lazy (readonly?) open of rpm database for upgrades. */
+    if (upgrade && rpmtsGetRdb(ts) == NULL && rpmtsGetDBMode(ts) != -1) {
+	if ((ec = rpmtsOpenDB(ts, rpmtsGetDBMode(ts))) != 0)
+	    goto exit;
+    }
+
+    p = rpmteNew(ts, h, TR_ADDED, key, relocs);
+    if (p == NULL) {
+	ec = 1;
+	goto exit;
+    }
+
     /* Check binary packages for redundancies in the set */
     if (!isSource) {
 	oc = findPos(ts, tscolor, h, upgrade);
@@ -281,12 +353,14 @@ int rpmtsAddInstallElement(rpmts ts, Header h,
 	    tsmem->order[oc] = rpmteFree(tsmem->order[oc]);
 	/* If newer NEVR was already added, we're done */
 	} else if (oc < 0) {
+	    p = rpmteFree(p);
 	    goto exit;
 	}
     }
 
     if (tsmem->addedPackages == NULL) {
-	tsmem->addedPackages = rpmalCreate(5, tscolor, rpmtsPrefColor(ts));
+	tsmem->addedPackages = rpmalCreate(5, rpmtsFlags(ts),
+					   tscolor, rpmtsPrefColor(ts));
     }
 
     if (oc >= tsmem->orderAlloced) {
@@ -295,30 +369,21 @@ int rpmtsAddInstallElement(rpmts ts, Header h,
 			tsmem->orderAlloced * sizeof(*tsmem->order));
     }
 
-    p = rpmteNew(ts, h, TR_ADDED, key, relocs);
 
     tsmem->order[oc] = p;
     if (oc == tsmem->orderCount) {
 	tsmem->orderCount++;
-	rpmcliPackagesTotal++;
     }
     
     rpmalAdd(tsmem->addedPackages, p);
 
-    /* If not upgrading or a source package, then we're done. */
-    if (!(upgrade & 0x1) || isSource)
-	goto exit;
-
-    /* Do lazy (readonly?) open of rpm database. */
-    if (rpmtsGetRdb(ts) == NULL && rpmtsGetDBMode(ts) != -1) {
-	if ((ec = rpmtsOpenDB(ts, rpmtsGetDBMode(ts))) != 0)
-	    goto exit;
+    /* Add erasure elements for old versions and obsoletions on upgrades */
+    /* XXX TODO: If either of these fails, we'd need to undo all additions */
+    if (upgrade) {
+	addUpgradeErasures(ts, tscolor, p, rpmteColor(p), h);
+	addObsoleteErasures(ts, tscolor, p);
     }
 
-    /* Add erasure elements for old versions and obsoletions */
-    addUpgradeErasures(ts, tsmem, tscolor, p, rpmteColor(p), h);
-    addObsoleteErasures(ts, tsmem, tscolor, p, rpmteColor(p));
-
 exit:
     return ec;
 }
@@ -333,21 +398,28 @@ static int rpmdbProvides(rpmts ts, depCache dcache, rpmds dep)
 {
     const char * Name = rpmdsN(dep);
     const char * DNEVR = rpmdsDNEVR(dep);
+    rpmTagVal deptag = rpmdsTagN(dep);
     int *cachedrc = NULL;
     rpmdbMatchIterator mi = NULL;
     Header h = NULL;
     int rc = 0;
+    /* pretrans deps are provided by current packages, don't prune erasures */
+    int prune = (rpmdsFlags(dep) & RPMSENSE_PRETRANS) ? 0 : 1;
 
     /* See if we already looked this up */
-    if (depCacheGetEntry(dcache, DNEVR, &cachedrc, NULL, NULL)) {
+    if (prune && depCacheGetEntry(dcache, DNEVR, &cachedrc, NULL, NULL)) {
 	rc = *cachedrc;
 	rpmdsNotify(dep, "(cached)", rc);
 	return rc;
     }
 
-    /* See if a filename dependency is a real file in some package */
-    if (Name[0] == '/') {
-	mi = rpmtsPrunedIterator(ts, RPMDBI_BASENAMES, Name);
+    /*
+     * See if a filename dependency is a real file in some package,
+     * taking file state into account: replaced, wrong colored and
+     * not installed files can not satisfy a dependency.
+     */
+    if (deptag != RPMTAG_OBSOLETENAME && Name[0] == '/') {
+	mi = rpmtsPrunedIterator(ts, RPMDBI_INSTFILENAMES, Name, prune);
 	while ((h = rpmdbNextIterator(mi)) != NULL) {
 	    rpmdsNotify(dep, "(db files)", rc);
 	    break;
@@ -357,9 +429,21 @@ static int rpmdbProvides(rpmts ts, depCache dcache, rpmds dep)
 
     /* Otherwise look in provides no matter what the dependency looks like */
     if (h == NULL) {
-	mi = rpmtsPrunedIterator(ts, RPMDBI_PROVIDENAME, Name);
+	/* Obsoletes use just name alone, everything else uses provides */
+	rpmTagVal dbtag = RPMDBI_PROVIDENAME;
+	if (deptag == RPMTAG_OBSOLETENAME)
+	    dbtag = RPMDBI_NAME;
+
+	mi = rpmtsPrunedIterator(ts, dbtag, Name, prune);
 	while ((h = rpmdbNextIterator(mi)) != NULL) {
-	    if (rpmdsAnyMatchesDep(h, dep, _rpmds_nopromote)) {
+	    int match;
+	    if (deptag == RPMTAG_OBSOLETENAME) {
+		match = rpmdsNVRMatchesDep(h, dep, _rpmds_nopromote);
+	    } else {
+		match = rpmdsMatchesDep(h, rpmdbGetIteratorFileNum(mi), dep,
+					_rpmds_nopromote);
+	    }
+	    if (match) {
 		rpmdsNotify(dep, "(db provides)", rc);
 		break;
 	    }
@@ -369,7 +453,9 @@ static int rpmdbProvides(rpmts ts, depCache dcache, rpmds dep)
     rc = (h != NULL) ? 0 : 1;
 
     /* Cache the relatively expensive rpmdb lookup results */
-    depCacheAddEntry(dcache, xstrdup(DNEVR), rc);
+    /* Caching the oddball non-pruned case would mess up other results */
+    if (prune)
+	depCacheAddEntry(dcache, xstrdup(DNEVR), rc);
     return rc;
 }
 
@@ -448,8 +534,14 @@ retry:
     }
 
 unsatisfied:
-    rc = 1;	/* dependency is unsatisfied */
-    rpmdsNotify(dep, NULL, rc);
+    if (dsflags & RPMSENSE_MISSINGOK) {
+	/* note the result, but missingok deps are never unsatisfied */
+	rpmdsNotify(dep, "(missingok)", 1);
+    } else {
+	/* dependency is unsatisfied */
+	rc = 1;
+	rpmdsNotify(dep, NULL, rc);
+    }
 
 exit:
     return rc;
@@ -485,7 +577,7 @@ static void checkInstDeps(rpmts ts, depCache dcache, rpmte te,
 			  rpmTag depTag, const char *dep)
 {
     Header h;
-    rpmdbMatchIterator mi = rpmtsPrunedIterator(ts, depTag, dep);
+    rpmdbMatchIterator mi = rpmtsPrunedIterator(ts, depTag, dep, 1);
 
     while ((h = rpmdbNextIterator(mi)) != NULL) {
 	char * pkgNEVRA = headerGetAsString(h, RPMTAG_NEVRA);
@@ -493,8 +585,8 @@ static void checkInstDeps(rpmts ts, depCache dcache, rpmte te,
 
 	checkDS(ts, dcache, te, pkgNEVRA, ds, dep, 0);
 
-	ds = rpmdsFree(ds);
-	pkgNEVRA = _free(pkgNEVRA);
+	rpmdsFree(ds);
+	free(pkgNEVRA);
     }
     rpmdbFreeIterator(mi);
 }
@@ -517,7 +609,7 @@ int rpmtsCheck(rpmts ts)
     }
 
     /* XXX FIXME: figure some kind of heuristic for the cache size */
-    dcache = depCacheCreate(5001, hashFunctionString, strcmp,
+    dcache = depCacheCreate(5001, rstrhash, strcmp,
 				     (depCacheFreeKey)rfree, NULL);
 
     /*
@@ -535,6 +627,8 @@ int rpmtsCheck(rpmts ts)
 		NULL, tscolor);
 	checkDS(ts, dcache, p, rpmteNEVRA(p), rpmteDS(p, RPMTAG_CONFLICTNAME),
 		NULL, tscolor);
+	checkDS(ts, dcache, p, rpmteNEVRA(p), rpmteDS(p, RPMTAG_OBSOLETENAME),
+		NULL, tscolor);
 
 	/* Check provides against conflicts in installed packages. */
 	while (rpmdsNext(provides) >= 0) {
@@ -548,7 +642,7 @@ int rpmtsCheck(rpmts ts)
 	/* Check package name (not provides!) against installed obsoletes */
 	checkInstDeps(ts, dcache, p, RPMTAG_OBSOLETENAME, rpmteN(p));
     }
-    pi = rpmtsiFree(pi);
+    rpmtsiFree(pi);
 
     /*
      * Look at the removed packages and make sure they aren't critical.
@@ -567,10 +661,11 @@ int rpmtsCheck(rpmts ts)
 	}
 
 	while (rpmfiNext(fi) >= 0) {
-	    checkInstDeps(ts, dcache, p, RPMTAG_REQUIRENAME, rpmfiFN(fi));
+	    if (RPMFILE_IS_INSTALLED(rpmfiFState(fi)))
+		checkInstDeps(ts, dcache, p, RPMTAG_REQUIRENAME, rpmfiFN(fi));
 	}
     }
-    pi = rpmtsiFree(pi);
+    rpmtsiFree(pi);
 
 exit:
     depCacheFree(dcache);
