@@ -68,6 +68,7 @@ struct rpmal_s {
     int delta;			/*!< Delta for pkg list reallocation. */
     int size;			/*!< No. of pkgs in list. */
     int alloced;		/*!< No. of pkgs allocated for list. */
+    rpmtransFlags tsflags;	/*!< Transaction control flags. */
     rpm_color_t tscolor;	/*!< Transaction color. */
     rpm_color_t prefcolor;	/*!< Transaction preferred color. */
 };
@@ -82,7 +83,8 @@ static void rpmalFreeIndex(rpmal al)
     al->fileHash = rpmalFileHashFree(al->fileHash);
 }
 
-rpmal rpmalCreate(int delta, rpm_color_t tscolor, rpm_color_t prefcolor)
+rpmal rpmalCreate(int delta, rpmtransFlags tsflags,
+		  rpm_color_t tscolor, rpm_color_t prefcolor)
 {
     rpmal al = xcalloc(1, sizeof(*al));
 
@@ -93,6 +95,7 @@ rpmal rpmalCreate(int delta, rpm_color_t tscolor, rpm_color_t prefcolor)
 
     al->providesHash = NULL;
     al->fileHash = NULL;
+    al->tsflags = tsflags;
     al->tscolor = tscolor;
     al->prefcolor = prefcolor;
 
@@ -120,11 +123,13 @@ rpmal rpmalFree(rpmal al)
     return NULL;
 }
 
-static unsigned int fileHash(struct fileNameEntry_s file){
-    return hashFunctionString(file.dirName) ^ hashFunctionString(file.baseName);
+static unsigned int fileHash(struct fileNameEntry_s file)
+{
+    return rstrhash(file.dirName) ^ rstrhash(file.baseName);
 }
 
-static int fileCompare(struct fileNameEntry_s one, struct fileNameEntry_s two) {
+static int fileCompare(struct fileNameEntry_s one, struct fileNameEntry_s two)
+{
     int rc = 0;
     rc = strcmp(one.dirName, two.dirName);
     if (!rc)
@@ -154,11 +159,14 @@ void rpmalDel(rpmal al, rpmte p)
     alp->p = NULL;
 }
 
-static void rpmalAddFiles(rpmal al, rpmalNum pkgNum, rpmfi fi){
+static void rpmalAddFiles(rpmal al, rpmalNum pkgNum, rpmfi fi)
+{
     struct fileNameEntry_s fileName;
     struct availableIndexEntry_s fileEntry;
     int i;
     rpm_color_t ficolor;
+    int skipdoc = (al->tsflags & RPMTRANS_FLAG_NODOCS);
+    int skipconf = (al->tsflags & RPMTRANS_FLAG_NOCONFIGS);
 
     fileEntry.pkgNum = pkgNum;
 
@@ -169,6 +177,12 @@ static void rpmalAddFiles(rpmal al, rpmalNum pkgNum, rpmfi fi){
         if (al->tscolor && ficolor && !(al->tscolor & ficolor))
             continue;
 
+	/* Ignore files that wont be installed */
+	if (skipdoc && (rpmfiFFlags(fi) & RPMFILE_DOC))
+	    continue;
+	if (skipconf && (rpmfiFFlags(fi) & RPMFILE_CONFIG))
+	    continue;
+
 	fileName.dirName = rpmfiDN(fi);
 	fileName.baseName = rpmfiBN(fi);
 
@@ -178,9 +192,11 @@ static void rpmalAddFiles(rpmal al, rpmalNum pkgNum, rpmfi fi){
     }
 }
 
-static void rpmalAddProvides(rpmal al, rpmalNum pkgNum, rpmds provides){
+static void rpmalAddProvides(rpmal al, rpmalNum pkgNum, rpmds provides)
+{
     struct availableIndexEntry_s indexEntry;
     rpm_color_t dscolor;
+    int skipconf = (al->tsflags & RPMTRANS_FLAG_NOCONFIGS);
 
     indexEntry.pkgNum = pkgNum;
 
@@ -191,6 +207,10 @@ static void rpmalAddProvides(rpmal al, rpmalNum pkgNum, rpmds provides){
         if (al->tscolor && dscolor && !(al->tscolor & dscolor))
             continue;
 
+	/* Ignore config() provides if the files wont be installed */
+	if (skipconf & (rpmdsFlags(provides) & RPMSENSE_CONFIG))
+	    continue;
+
 	indexEntry.entryIx = rpmdsIx(provides);
 	rpmalProvidesHashAddEntry(al->providesHash, rpmdsN(provides), indexEntry);
     }
@@ -240,7 +260,7 @@ static void rpmalMakeIndex(rpmal al)
 	    fileCnt += rpmfiFC(alp->fi);
     }
 
-    al->providesHash = rpmalProvidesHashCreate(providesCnt/4+128, hashFunctionString,
+    al->providesHash = rpmalProvidesHashCreate(providesCnt/4+128, rstrhash,
 					       strcmp, NULL, NULL);
     al->fileHash = rpmalFileHashCreate(fileCnt/4+128, fileHash, fileCompare,
 				       NULL, NULL);
@@ -304,6 +324,7 @@ static rpmte * rpmalAllSatisfiesDepend(const rpmal al, const rpmds ds)
     const char * name;
     availableIndexEntry result;
     int resultCnt;
+    int obsolete;
 
     availablePackage alp;
     int rc;
@@ -314,7 +335,8 @@ static rpmte * rpmalAllSatisfiesDepend(const rpmal al, const rpmds ds)
     if (al->providesHash == NULL && al->fileHash == NULL)
 	rpmalMakeIndex(al);
 
-    if (*name == '/') {
+    obsolete = (rpmdsTagN(ds) == RPMTAG_OBSOLETENAME);
+    if (!obsolete && *name == '/') {
 	/* First, look for files "contained" in package ... */
 	ret = rpmalAllFileSatisfiesDepend(al, ds);
 	if (ret != NULL && *ret != NULL)
@@ -335,6 +357,11 @@ static rpmte * rpmalAllSatisfiesDepend(const rpmal al, const rpmds ds)
 	if (alp->p == NULL) // deleted
 	    continue;
 	(void) rpmdsSetIx(alp->provides, result[i].entryIx);
+
+	/* Obsoletes are on package name, filter out other provide matches */
+	if (obsolete && !rstreq(rpmdsN(alp->provides), rpmteN(alp->p)))
+	    continue;
+
 	rc = 0;
 	if (rpmdsIx(alp->provides) >= 0)
 	    rc = rpmdsCompare(alp->provides, ds);
