@@ -15,7 +15,6 @@
 #include "lib/rpmfi_internal.h"
 #include "lib/rpmte_internal.h"	/* relocations */
 #include "lib/cpio.h"	/* XXX CPIO_FOO */
-#include "lib/fsm.h"	/* XXX newFSM() */
 
 #include "debug.h"
 
@@ -52,7 +51,7 @@ static scidx_t strcachePut(strcache cache, const char *str)
 	/* blow up on index wraparound */
 	assert((scidx_t)(cache->num + 1) > cache->num);
 	cache->uniq = xrealloc(cache->uniq, 
-				sizeof(cache->uniq) * (cache->num+1));
+				sizeof(*cache->uniq) * (cache->num+1));
 	cache->uniq[cache->num] = xstrdup(str);
 	ret = cache->num;
 	cache->num++;
@@ -181,30 +180,13 @@ const char * rpmfiDNIndex(rpmfi fi, int jx)
     return DN;
 }
 
-const char * rpmfiFNIndex(rpmfi fi, int ix)
+char * rpmfiFNIndex(rpmfi fi, int ix)
 {
-    const char * FN = "";
-
+    char *fn = NULL;
     if (fi != NULL && ix >= 0 && ix < fi->fc) {
-	char * t;
-	if (fi->fn == NULL) {
-	    size_t dnlmax = 0, bnlmax = 0, len;
-	    for (int i = 0; i < fi->dc; i++) {
-		if ((len = strlen(fi->dnl[i])) > dnlmax)
-		    dnlmax = len;
-	    }
-	    for (int i = 0; i < fi->fc; i++) {
-		if ((len = strlen(fi->bnl[i])) > bnlmax)
-		    bnlmax = len;
-	    }
-	    fi->fn = xmalloc(dnlmax + bnlmax + 1);
-	}
-	FN = t = fi->fn;
-	*t = '\0';
-	t = stpcpy(t, fi->dnl[fi->dil[ix]]);
-	t = stpcpy(t, fi->bnl[ix]);
+	fn = rstrscat(NULL, fi->dnl[fi->dil[ix]], fi->bnl[ix], NULL);
     }
-    return FN;
+    return fn;
 }
 
 rpmfileAttrs rpmfiFFlagsIndex(rpmfi fi, int ix)
@@ -558,19 +540,25 @@ rpmFileTypes rpmfiWhatis(rpm_mode_t mode)
     return REG;
 }
 
-int rpmfiCompare(const rpmfi afi, const rpmfi bfi)
+int rpmfiCompareIndex(rpmfi afi, int aix, rpmfi bfi, int bix)
 {
-    rpmFileTypes awhat = rpmfiWhatis(rpmfiFMode(afi));
-    rpmFileTypes bwhat = rpmfiWhatis(rpmfiFMode(bfi));
+    mode_t amode = rpmfiFModeIndex(afi, aix);
+    mode_t bmode = rpmfiFModeIndex(bfi, bix);
+    rpmFileTypes awhat = rpmfiWhatis(amode);
 
-    if ((rpmfiFFlags(afi) & RPMFILE_GHOST) ||
-	(rpmfiFFlags(bfi) & RPMFILE_GHOST)) return 0;
+    if ((rpmfiFFlagsIndex(afi, aix) & RPMFILE_GHOST) ||
+	(rpmfiFFlagsIndex(bfi, bix) & RPMFILE_GHOST)) return 0;
 
-    if (awhat != bwhat) return 1;
+    if (amode != bmode) return 1;
+
+    if (!rstreq(rpmfiFUserIndex(afi, aix), rpmfiFUserIndex(bfi, bix)))
+	return 1;
+    if (!rstreq(rpmfiFGroupIndex(afi, aix), rpmfiFGroupIndex(bfi, bix)))
+	return 1;
 
     if (awhat == LINK) {
-	const char * alink = rpmfiFLink(afi);
-	const char * blink = rpmfiFLink(bfi);
+	const char * alink = rpmfiFLinkIndex(afi, aix);
+	const char * blink = rpmfiFLinkIndex(bfi, bix);
 	if (alink == blink) return 0;
 	if (alink == NULL) return 1;
 	if (blink == NULL) return -1;
@@ -578,27 +566,33 @@ int rpmfiCompare(const rpmfi afi, const rpmfi bfi)
     } else if (awhat == REG) {
 	size_t adiglen, bdiglen;
 	int aalgo, balgo;
-	const unsigned char * adigest = rpmfiFDigest(afi, &aalgo, &adiglen);
-	const unsigned char * bdigest = rpmfiFDigest(bfi, &balgo, &bdiglen);
+	const unsigned char * adigest, * bdigest;
+	adigest = rpmfiFDigestIndex(afi, aix, &aalgo, &adiglen);
+	bdigest = rpmfiFDigestIndex(bfi, bix, &balgo, &bdiglen);
 	if (adigest == bdigest) return 0;
 	if (adigest == NULL) return 1;
 	if (bdigest == NULL) return -1;
 	/* can't meaningfully compare different hash types */
 	if (aalgo != balgo || adiglen != bdiglen) return -1;
 	return memcmp(adigest, bdigest, adiglen);
+    } else if (awhat == CDEV || awhat == BDEV) {
+	if (rpmfiFRdevIndex(afi, aix) != rpmfiFRdevIndex(bfi, bix))
+	    return 1;
     }
 
     return 0;
 }
 
-rpmFileAction rpmfiDecideFate(const rpmfi ofi, rpmfi nfi, int skipMissing)
+rpmFileAction rpmfiDecideFateIndex(rpmfi ofi, int oix, rpmfi nfi, int nix,
+				   int skipMissing)
 {
-    const char * fn = rpmfiFN(nfi);
-    rpmfileAttrs newFlags = rpmfiFFlags(nfi);
+    char * fn = rpmfiFNIndex(nfi, nix);
+    rpmfileAttrs newFlags = rpmfiFFlagsIndex(nfi, nix);
     char buffer[1024];
     rpmFileTypes dbWhat, newWhat, diskWhat;
     struct stat sb;
     int save = (newFlags & RPMFILE_NOREPLACE) ? FA_ALTNAME : FA_SAVE;
+    int action = FA_CREATE; /* assume we can create */
 
     if (lstat(fn, &sb)) {
 	/*
@@ -608,31 +602,33 @@ rpmFileAction rpmfiDecideFate(const rpmfi ofi, rpmfi nfi, int skipMissing)
 	if (skipMissing && (newFlags & RPMFILE_MISSINGOK)) {
 	    rpmlog(RPMLOG_DEBUG, "%s skipped due to missingok flag\n",
 			fn);
-	    return FA_SKIP;
-	} else {
-	    return FA_CREATE;
+	    action = FA_SKIP;
 	}
+	goto exit;
     }
 
     diskWhat = rpmfiWhatis((rpm_mode_t)sb.st_mode);
-    dbWhat = rpmfiWhatis(rpmfiFMode(ofi));
-    newWhat = rpmfiWhatis(rpmfiFMode(nfi));
+    dbWhat = rpmfiWhatis(rpmfiFModeIndex(ofi, oix));
+    newWhat = rpmfiWhatis(rpmfiFModeIndex(nfi, nix));
 
     /*
      * RPM >= 2.3.10 shouldn't create config directories -- we'll ignore
      * them in older packages as well.
      */
     if (newWhat == XDIR)
-	return FA_CREATE;
+	goto exit;
 
-    if (diskWhat != newWhat && dbWhat != REG && dbWhat != LINK)
-	return save;
-    else if (newWhat != dbWhat && diskWhat != dbWhat)
-	return save;
-    else if (dbWhat != newWhat)
-	return FA_CREATE;
-    else if (dbWhat != LINK && dbWhat != REG)
-	return FA_CREATE;
+    if (diskWhat != newWhat && dbWhat != REG && dbWhat != LINK) {
+	action = save;
+	goto exit;
+    } else if (newWhat != dbWhat && diskWhat != dbWhat) {
+	action = save;
+	goto exit;
+    } else if (dbWhat != newWhat) {
+	goto exit;
+    } else if (dbWhat != LINK && dbWhat != REG) {
+	goto exit;
+    }
 
     /*
      * This order matters - we'd prefer to CREATE the file if at all
@@ -643,83 +639,105 @@ rpmFileAction rpmfiDecideFate(const rpmfi ofi, rpmfi nfi, int skipMissing)
 	int oalgo, nalgo;
 	size_t odiglen, ndiglen;
 	const unsigned char * odigest, * ndigest;
-	odigest = rpmfiFDigest(ofi, &oalgo, &odiglen);
+	odigest = rpmfiFDigestIndex(ofi, oix, &oalgo, &odiglen);
 	if (diskWhat == REG) {
-	    if (rpmDoDigest(oalgo, fn, 0, 
-		(unsigned char *)buffer, NULL))
-	        return FA_CREATE;	/* assume file has been removed */
+	    if (rpmDoDigest(oalgo, fn, 0, (unsigned char *)buffer, NULL))
+	        goto exit;	/* assume file has been removed */
 	    if (odigest && !memcmp(odigest, buffer, odiglen))
-	        return FA_CREATE;	/* unmodified config file, replace. */
+	        goto exit;	/* unmodified config file, replace. */
 	}
-	ndigest = rpmfiFDigest(nfi, &nalgo, &ndiglen);
+	ndigest = rpmfiFDigestIndex(nfi, nix, &nalgo, &ndiglen);
 	/* Can't compare different hash types, backup to avoid data loss */
-	if (oalgo != nalgo || odiglen != ndiglen)
-	    return save;
-	if (odigest && ndigest && !memcmp(odigest, ndigest, odiglen))
-	    return FA_SKIP;	/* identical file, don't bother. */
+	if (oalgo != nalgo || odiglen != ndiglen) {
+	    action = save;
+	    goto exit;
+	}
+	if (odigest && ndigest && !memcmp(odigest, ndigest, odiglen)) {
+	    action = FA_SKIP;	/* identical file, don't bother. */
+	    goto exit;
+	}
+	/* ... but otherwise backup will be needed */
+	action = save;
     } else /* dbWhat == LINK */ {
 	const char * oFLink, * nFLink;
-	oFLink = rpmfiFLink(ofi);
+	oFLink = rpmfiFLinkIndex(ofi, oix);
 	if (diskWhat == LINK) {
-	    if (readlink(fn, buffer, sizeof(buffer) - 1) == -1)
-		return FA_CREATE;	/* assume file has been removed */
+	    ssize_t link_len = readlink(fn, buffer, sizeof(buffer) - 1);
+	    if (link_len == -1)
+		goto exit;	/* assume file has been removed */
+	    buffer[link_len] = '\0';
 	    if (oFLink && rstreq(oFLink, buffer))
-		return FA_CREATE;	/* unmodified config file, replace. */
+		goto exit;	/* unmodified config file, replace. */
 	}
-	nFLink = rpmfiFLink(nfi);
-	if (oFLink && nFLink && rstreq(oFLink, nFLink))
-	    return FA_SKIP;	/* identical file, don't bother. */
+	nFLink = rpmfiFLinkIndex(nfi, nix);
+	if (oFLink && nFLink && rstreq(oFLink, nFLink)) {
+	    action = FA_SKIP;	/* identical file, don't bother. */
+	    goto exit;
+	}
+	/* ... but otherwise backup will be needed */
+	action = save;
     }
 
-    /*
-     * The config file on the disk has been modified, but
-     * the ones in the two packages are different. It would
-     * be nice if RPM was smart enough to at least try and
-     * merge the difference ala CVS, but...
-     */
-    return save;
+exit:
+    free(fn);
+    return action;
 }
 
-int rpmfiConfigConflict(const rpmfi fi)
+int rpmfiConfigConflictIndex(rpmfi fi, int ix)
 {
-    const char * fn = rpmfiFN(fi);
-    rpmfileAttrs flags = rpmfiFFlags(fi);
+    char * fn = NULL;
+    rpmfileAttrs flags = rpmfiFFlagsIndex(fi, ix);
     char buffer[1024];
     rpmFileTypes newWhat, diskWhat;
     struct stat sb;
+    int rc = 0;
 
-    if (!(flags & RPMFILE_CONFIG) || lstat(fn, &sb)) {
+    if (!(flags & RPMFILE_CONFIG))
 	return 0;
-    }
-
-    diskWhat = rpmfiWhatis((rpm_mode_t)sb.st_mode);
-    newWhat = rpmfiWhatis(rpmfiFMode(fi));
 
+    /* Only links and regular files can be %config, this is kinda moot */
+    /* XXX: Why are we returning 1 here? */
+    newWhat = rpmfiWhatis(rpmfiFModeIndex(fi, ix));
     if (newWhat != LINK && newWhat != REG)
 	return 1;
 
-    if (diskWhat != newWhat)
-	return 1;
+    /* If it's not on disk, there's nothing to be saved */
+    fn = rpmfiFNIndex(fi, ix);
+    if (lstat(fn, &sb))
+	goto exit;
+
+    /* Files of different types obviously are not identical */
+    diskWhat = rpmfiWhatis((rpm_mode_t)sb.st_mode);
+    if (diskWhat != newWhat) {
+	rc = 1;
+	goto exit;
+    }
     
     memset(buffer, 0, sizeof(buffer));
     if (newWhat == REG) {
 	int algo;
 	size_t diglen;
-	const unsigned char *ndigest = rpmfiFDigest(fi, &algo, &diglen);
+	const unsigned char *ndigest = rpmfiFDigestIndex(fi,ix, &algo, &diglen);
 	if (rpmDoDigest(algo, fn, 0, (unsigned char *)buffer, NULL))
-	    return 0;	/* assume file has been removed */
+	    goto exit;	/* assume file has been removed */
 	if (ndigest && !memcmp(ndigest, buffer, diglen))
-	    return 0;	/* unmodified config file */
+	    goto exit;	/* unmodified config file */
     } else /* newWhat == LINK */ {
 	const char * nFLink;
-	if (readlink(fn, buffer, sizeof(buffer) - 1) == -1)
-	    return 0;	/* assume file has been removed */
-	nFLink = rpmfiFLink(fi);
+	ssize_t link_len = readlink(fn, buffer, sizeof(buffer) - 1);
+	if (link_len == -1)
+	    goto exit;	/* assume file has been removed */
+	buffer[link_len] = '\0';
+	nFLink = rpmfiFLinkIndex(fi, ix);
 	if (nFLink && rstreq(nFLink, buffer))
-	    return 0;	/* unmodified config file */
+	    goto exit;	/* unmodified config file */
     }
 
-    return 1;
+    rc = 1;
+
+exit:
+    free(fn);
+    return rc;
 }
 
 static char **duparray(char ** src, int size)
@@ -783,7 +801,7 @@ static int addPrefixes(Header h, rpmRelocation *relocations, int numRelocations)
     if (numActual) {
 	headerPutStringArray(h, RPMTAG_INSTPREFIXES, actualRelocations, numActual);
     }
-    actualRelocations = _free(actualRelocations);
+    free(actualRelocations);
     return numActual;
 }
 
@@ -1081,8 +1099,6 @@ rpmfi rpmfiFree(rpmfi fi)
 	}
     }
 
-    fi->fsm = freeFSM(fi->fsm);
-
     fi->fn = _free(fi->fn);
     fi->apath = _free(fi->apath);
 
@@ -1120,38 +1136,18 @@ static scidx_t *cacheTag(strcache cache, Header h, rpmTag tag)
 
 rpmfi rpmfiNew(const rpmts ts, Header h, rpmTagVal tagN, rpmfiFlags flags)
 {
-    rpmfi fi = NULL;
-    rpm_loff_t *asize = NULL;
+    rpmfi fi = xcalloc(1, sizeof(*fi)); 
     unsigned char * t;
-    int isBuild, isSource;
     struct rpmtd_s fdigests, digalgo;
     struct rpmtd_s td;
     headerGetFlags scareFlags = (flags & RPMFI_KEEPHEADER) ? 
 				HEADERGET_MINMEM : HEADERGET_ALLOC;
     headerGetFlags defFlags = HEADERGET_ALLOC;
 
-    fi = xcalloc(1, sizeof(*fi));
-    if (fi == NULL)	/* XXX can't happen */
-	goto exit;
-
     fi->magic = RPMFIMAGIC;
     fi->i = -1;
 
     fi->fiflags = flags;
-    fi->scareFlags = scareFlags;
-
-    if (headerGet(h, RPMTAG_LONGARCHIVESIZE, &td, HEADERGET_EXT)) {
-	asize = rpmtdGetUint64(&td);
-    }
-    /* 0 means unknown */
-    fi->archiveSize = asize ? *asize : 0;
-    rpmtdFreeData(&td);
-    
-    /* Archive size is not set when this gets called from build */
-    isBuild = (asize == NULL);
-    isSource = headerIsSource(h);
-    if (isBuild) fi->fiflags |= RPMFI_ISBUILD;
-    if (isSource) fi->fiflags |= RPMFI_ISSOURCE;
 
     _hgfi(h, RPMTAG_BASENAMES, &td, defFlags, fi->bnl);
     fi->fc = rpmtdCount(&td);
@@ -1162,6 +1158,17 @@ rpmfi rpmfiNew(const rpmts ts, Header h, rpmTagVal tagN, rpmfiFlags flags)
     _hgfi(h, RPMTAG_DIRNAMES, &td, defFlags, fi->dnl);
     fi->dc = rpmtdCount(&td);
     _hgfi(h, RPMTAG_DIRINDEXES, &td, scareFlags, fi->dil);
+
+    /* Is our filename triplet sane? */
+    if (fi->dc == 0 || fi->dc > fi->fc || rpmtdCount(&td) != fi->fc)
+	goto errxit;
+
+    for (rpm_count_t i = 0; i < fi->fc; i++) {
+	if (fi->dil[i] >= fi->fc)
+	    goto errxit;
+    }
+
+    /* XXX TODO: all these should be sanity checked, ugh... */
     if (!(flags & RPMFI_NOFILEMODES))
 	_hgfi(h, RPMTAG_FILEMODES, &td, scareFlags, fi->fmodes);
     if (!(flags & RPMFI_NOFILEFLAGS))
@@ -1256,25 +1263,29 @@ exit:
 
     /* FIX: rpmfi null annotations */
     return rpmfiLink(fi);
+
+errxit:
+    rpmfiFree(fi);
+    return NULL;
 }
 
-void rpmfiSetFReplacedSize(rpmfi fi, rpm_loff_t newsize)
+void rpmfiSetFReplacedSizeIndex(rpmfi fi, int ix, rpm_loff_t newsize)
 {
-    if (fi != NULL && fi->i >= 0 && fi->i < fi->fc) {
+    if (fi != NULL && ix >= 0 && ix < fi->fc) {
 	if (fi->replacedSizes == NULL) {
 	    fi->replacedSizes = xcalloc(fi->fc, sizeof(*fi->replacedSizes));
 	}
 	/* XXX watch out, replacedSizes is not rpm_loff_t (yet) */
-	fi->replacedSizes[fi->i] = (rpm_off_t) newsize;
+	fi->replacedSizes[ix] = (rpm_off_t) newsize;
     }
 }
 
-rpm_loff_t rpmfiFReplacedSize(rpmfi fi)
+rpm_loff_t rpmfiFReplacedSizeIndex(rpmfi fi, int ix)
 {
     rpm_loff_t rsize = 0;
-    if (fi != NULL && fi->i >= 0 && fi->i < fi->fc) {
+    if (fi != NULL && ix >= 0 && ix < fi->fc) {
 	if (fi->replacedSizes) {
-	    rsize = fi->replacedSizes[fi->i];
+	    rsize = fi->replacedSizes[ix];
 	}
     }
     return rsize;
@@ -1288,28 +1299,6 @@ void rpmfiFpLookup(rpmfi fi, fingerPrintCache fpc)
     fpLookupList(fpc, fi->dnl, fi->bnl, fi->dil, fi->fc, fi->fps);
 }
 
-FSM_t rpmfiFSM(rpmfi fi)
-{
-    if (fi != NULL && fi->fsm == NULL) {
-    	cpioMapFlags mapflags;
-	/* Figure out mapflags: 
-	 * - path, mode, uid and gid are used by everything
-	 * - all binary packages get SBIT_CHECK set
-	 * - if archive size is not known, we're only building this package,
-	 *   different rules apply 
-	 */
-	mapflags = CPIO_MAP_PATH | CPIO_MAP_MODE | CPIO_MAP_UID | CPIO_MAP_GID;
-	if (fi->fiflags & RPMFI_ISBUILD) {
-	    mapflags |= CPIO_MAP_TYPE;
-	    if (fi->fiflags & RPMFI_ISSOURCE) mapflags |= CPIO_FOLLOW_SYMLINKS;
-	} else {
-	    if (!(fi->fiflags & RPMFI_ISSOURCE)) mapflags |= CPIO_SBIT_CHECK;
-	}
-	fi->fsm = newFSM(mapflags);
-    }
-    return (fi != NULL) ? fi->fsm : NULL;
-}
-
 /* 
  * Generate iterator accessors function wrappers, these do nothing but
  * call the corresponding rpmfiFooIndex(fi, fi->[ij])
@@ -1320,7 +1309,6 @@ FSM_t rpmfiFSM(rpmfi fi)
 
 RPMFI_ITERFUNC(const char *, BN, i)
 RPMFI_ITERFUNC(const char *, DN, j)
-RPMFI_ITERFUNC(const char *, FN, i)
 RPMFI_ITERFUNC(const char *, FLink, i)
 RPMFI_ITERFUNC(const char *, FUser, i)
 RPMFI_ITERFUNC(const char *, FGroup, i)
@@ -1338,6 +1326,18 @@ RPMFI_ITERFUNC(rpm_loff_t, FSize, i)
 RPMFI_ITERFUNC(rpm_color_t, FColor, i)
 RPMFI_ITERFUNC(uint32_t, FNlink, i)
 
+const char * rpmfiFN(rpmfi fi)
+{
+    const char *fn = ""; /* preserve behavior on errors */
+    if (fi != NULL) {
+	free(fi->fn);
+	fi->fn = rpmfiFNIndex(fi, fi->i);
+	if (fi->fn != NULL)
+	    fn = fi->fn;
+    }
+    return fn;
+}
+
 const unsigned char * rpmfiFDigest(rpmfi fi, int *algo, size_t *len)
 {
     return rpmfiFDigestIndex(fi, fi ? fi->i : -1, algo, len);
@@ -1347,3 +1347,20 @@ uint32_t rpmfiFDepends(rpmfi fi, const uint32_t ** fddictp)
 {
     return rpmfiFDependsIndex(fi,  fi ? fi->i : -1, fddictp);
 }
+
+int rpmfiCompare(const rpmfi afi, const rpmfi bfi)
+{
+    return rpmfiCompareIndex(afi, afi ? afi->i : -1, bfi, bfi ? bfi->i : -1);
+}
+
+rpmFileAction rpmfiDecideFate(const rpmfi ofi, rpmfi nfi, int skipMissing)
+{
+    return rpmfiDecideFateIndex(ofi, ofi ? ofi->i : -1,
+				nfi, nfi ? nfi->i : -1,
+				skipMissing);
+}
+
+int rpmfiConfigConflict(const rpmfi fi)
+{
+    return rpmfiConfigConflictIndex(fi, fi ? fi->i : -1);
+}
