@@ -4,7 +4,8 @@
 
 #include "system.h"
 
-#include <rpmlib.h>
+#include <rpm/rpmlib.h>		/* rpmvercmp proto */
+#include <rpm/rpmstring.h>
 
 #include "debug.h"
 
@@ -15,16 +16,14 @@
 int rpmvercmp(const char * a, const char * b)
 {
     char oldch1, oldch2;
-    char * str1, * str2;
+    char abuf[strlen(a)+1], bbuf[strlen(b)+1];
+    char *str1 = abuf, *str2 = bbuf;
     char * one, * two;
     int rc;
     int isnum;
 
     /* easy comparison to see if versions are identical */
-    if (!strcmp(a, b)) return 0;
-
-    str1 = alloca(strlen(a) + 1);
-    str2 = alloca(strlen(b) + 1);
+    if (rstreq(a, b)) return 0;
 
     strcpy(str1, a);
     strcpy(str2, b);
@@ -33,11 +32,21 @@ int rpmvercmp(const char * a, const char * b)
     two = str2;
 
     /* loop through each version segment of str1 and str2 and compare them */
-    /*@-branchstate@*/
-/*@-boundsread@*/
-    while (*one && *two) {
-	while (*one && !xisalnum(*one)) one++;
-	while (*two && !xisalnum(*two)) two++;
+    while (*one || *two) {
+	while (*one && !risalnum(*one) && *one != '~') one++;
+	while (*two && !risalnum(*two) && *two != '~') two++;
+
+	/* handle the tilde separator, it sorts before everthing else */
+	if (*one == '~' || *two == '~') {
+	    if (*one != '~') return 1;
+	    if (*two != '~') return -1;
+	    one++;
+	    two++;
+	    continue;
+	}
+
+	/* If we ran to the end of either, we are finished with the loop */
+	if (!(*one && *two)) break;
 
 	str1 = one;
 	str2 = two;
@@ -45,28 +54,30 @@ int rpmvercmp(const char * a, const char * b)
 	/* grab first completely alpha or completely numeric segment */
 	/* leave one and two pointing to the start of the alpha or numeric */
 	/* segment and walk str1 and str2 to end of segment */
-	if (xisdigit(*str1)) {
-	    while (*str1 && xisdigit(*str1)) str1++;
-	    while (*str2 && xisdigit(*str2)) str2++;
+	if (risdigit(*str1)) {
+	    while (*str1 && risdigit(*str1)) str1++;
+	    while (*str2 && risdigit(*str2)) str2++;
 	    isnum = 1;
 	} else {
-	    while (*str1 && xisalpha(*str1)) str1++;
-	    while (*str2 && xisalpha(*str2)) str2++;
+	    while (*str1 && risalpha(*str1)) str1++;
+	    while (*str2 && risalpha(*str2)) str2++;
 	    isnum = 0;
 	}
 
 	/* save character at the end of the alpha or numeric segment */
 	/* so that they can be restored after the comparison */
-/*@-boundswrite@*/
 	oldch1 = *str1;
 	*str1 = '\0';
 	oldch2 = *str2;
 	*str2 = '\0';
-/*@=boundswrite@*/
+
+	/* this cannot happen, as we previously tested to make sure that */
+	/* the first string has a non-null segment */
+	if (one == str1) return -1;	/* arbitrary */
 
 	/* take care of the case where the two version segments are */
 	/* different types: one numeric, the other alpha (i.e. empty) */
-	if (one == str1) return -1;	/* arbitrary */
+	/* numeric segments are always newer than alpha segments */
 	/* XXX See patch #60884 (and details) from bugzilla #50977. */
 	if (two == str2) return (isnum ? 1 : -1);
 
@@ -89,26 +100,41 @@ int rpmvercmp(const char * a, const char * b)
 	/* if they are equal because there might be more segments to */
 	/* compare */
 	rc = strcmp(one, two);
-	if (rc) return rc;
+	if (rc) return (rc < 1 ? -1 : 1);
 
 	/* restore character that was replaced by null above */
-/*@-boundswrite@*/
 	*str1 = oldch1;
 	one = str1;
 	*str2 = oldch2;
 	two = str2;
-/*@=boundswrite@*/
     }
-    /*@=branchstate@*/
-/*@=boundsread@*/
 
     /* this catches the case where all numeric and alpha segments have */
     /* compared identically but the segment sepparating characters were */
     /* different */
-/*@-boundsread@*/
     if ((!*one) && (!*two)) return 0;
 
     /* whichever version still has characters left over wins */
     if (!*one) return -1; else return 1;
-/*@=boundsread@*/
+}
+
+int rpmVersionCompare(Header first, Header second)
+{
+    /* Missing epoch becomes zero here, which is what we want */
+    uint32_t epochOne = headerGetNumber(first, RPMTAG_EPOCH);
+    uint32_t epochTwo = headerGetNumber(second, RPMTAG_EPOCH);
+    int rc;
+
+    if (epochOne < epochTwo)
+	return -1;
+    else if (epochOne > epochTwo)
+	return 1;
+
+    rc = rpmvercmp(headerGetString(first, RPMTAG_VERSION),
+		   headerGetString(second, RPMTAG_VERSION));
+    if (rc)
+	return rc;
+
+    return rpmvercmp(headerGetString(first, RPMTAG_RELEASE),
+		     headerGetString(second, RPMTAG_RELEASE));
 }
